(*
 * This file is part of Barista.
 * Copyright (C) 2007-2012 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)


open Utils

(* Types *)

type entry =
  | Directory of string
  | Archive of Zip.in_file
  | Big_archive of string * Zip.in_file * (string list)

type t = entry list


(* Exception *)

BARISTA_ERROR =
  | Unable_to_open_archive of (s : string) ->
      Printf.sprintf "unable to open archive %S" s
  | Does_not_exist of (s : string) ->
      Printf.sprintf "%S does not exist" s
  | Class_not_found of (s : string) ->
      Printf.sprintf "class %S not found" s


(* Constructor *)

let special_archive filename prefix =
  let len_prefix = String.length prefix in
  try
    let zip = Zip.open_in filename in
    let prefixes = ref [] in
    List.iter
      (fun e ->
        let name = e.Zip.filename in
        let len = String.length name in
        if (len > len_prefix)
            && ((String.sub name 0 len_prefix) = prefix) then begin
              try
                let index = String.index_from name len_prefix '/' in
                let jar_name = String.sub name len_prefix (index - len_prefix) in
                if not (List.mem jar_name !prefixes) then
                  prefixes := jar_name :: !prefixes
              with Not_found -> ()
            end)
      (Zip.entries zip);
    Big_archive (prefix, zip, !prefixes)
  with _ -> fail (Unable_to_open_archive filename)

let sym_archive_prefix = "META-INF/sym/"

let web_archive_prefix = "WEB-INF/classes"

let make_entry s =
  try
    if Sys.is_directory s then
      let last = (String.length s) - 1 in
      Directory (if s.[last] = '/' then s else s ^ "/")
    else if Sys.file_exists s then begin
      if Filename.check_suffix s ".sym" then
        special_archive s sym_archive_prefix
      else if Filename.check_suffix s ".war" then
        (try
          Big_archive (web_archive_prefix, Zip.open_in s, [""])
        with _ -> fail (Unable_to_open_archive s))
      else
        (try
          Archive (Zip.open_in s)
        with _ -> fail (Unable_to_open_archive s))
    end else
      fail (Does_not_exist s)
  with Sys_error _ -> fail (Does_not_exist s)

let map_elements l =
  List.map make_entry l

let make_of_string ?(separator=":") s =
  let l = string_split separator s in
  map_elements l

let make_of_list l =
  map_elements l

let make () =
  let jdk_classes =
    try
      let home =
        if Sys.file_exists "/usr/libexec/java_home" then begin
          let temp = Filename.temp_file "barista" "javahome" in
          let code = Sys.command ("/usr/libexec/java_home 2> /dev/null > " ^ temp) in
          if code <> 0 then raise Not_found;
          let chan = open_in temp in
          let line = input_line chan in
          close_in_noerr chan;
          line
        end else
          Sys.getenv "JAVA_HOME" in
      let jdk_archive = home ^ "/lib/ct.sym" in
      [make_entry jdk_archive]
    with _ -> [] in
  let s = try Sys.getenv "CLASSPATH" with Not_found -> "." in
  jdk_classes @ (make_of_string s)

let append ?(separator=":") s l =
  let l' = make_of_string ~separator s in
  l @ l'

let prepend ?(separator=":") s l =
  let l' = make_of_string ~separator s in
  l' @ l


(* Functions *)

let open_entry z s =
  let entry = Zip.find_entry z s in
  let data = Zip.read_entry z entry in
  InputStream.make_of_string data

let rec lookup_with_prefix p z s = function
  | prefix :: tl ->
      (try
        let full_name =
          Printf.sprintf "%s%s/%s"
            p
            prefix
            s in
        open_entry z full_name
      with _ -> lookup_with_prefix p z s tl)
  | [] -> fail (Class_not_found s)

let rec lookup cp s =
  match cp with
  | (Directory d) :: tl ->
      (try
        InputStream.make_of_channel (open_in (d ^ s))
      with _ -> lookup tl s)
  | (Archive z) :: tl ->
      (try
        open_entry z s
      with _ -> lookup tl s)
  | (Big_archive (p, z, pl)) :: tl ->
      (try
        lookup_with_prefix p z s pl
      with _ -> lookup tl s)
  | [] -> fail (Class_not_found s)

let search_stream cp s =
  let rec search () =
    try
      lookup cp s
    with _ ->
      try
        s.[String.rindex s '/'] <- '$';
        search ()
      with _ -> fail (Class_not_found s) in
  search ()

let open_stream cp s =
  let s' = (string_replace '.' '/' s) ^ ".class" in
  try
    search_stream cp s'
  with _ -> fail (Class_not_found s)

let close cp =
  let close_entry = function
    | Directory _ -> ()
    | Archive z | Big_archive (_, z, _) ->
        try Zip.close_in z with _ -> () in
  List.iter close_entry cp
